"""
The control chart constants for the X Bar R control chart.
"""

from SPC import ControlChartConstants

class XbarRControlChartConstants(ControlChartConstants):
    def __init__(self):
        self._constants = {
            2: [1.880, 1.128, 0, 3.267],
            3: [1.023, 1.693, 0, 2.574],
            4: [0.729, 2.059, 0, 2.282],
            5: [0.577, 2.326, 0, 2.114],
            6: [0.483, 2.534, 0, 2.004],
            7: [0.419, 2.704, 0.076, 1.924],
            8: [0.373, 2.847, 0.136, 0.1864],
            9: [0.337, 2.970, 0.184, 1.816],
            10: [0.308, 3.078, 0.223, 1.777],
            15: [0.223, 3.472, 0.347, 1.653],
            25: [0.153, 3.931, 0.459, 1.541]
            # Add more values as needed n: [A2, d2, D3, D4]
        }

    def get_constant(self, n, idx):
        """ Returns the control chart constant.

            :param n: The subgroup size.
            :param idx: The index.
        """
        try:
            value= self._constants[n][idx]
            return value
        except KeyError:
            raise ValueError("invalid parameter value(s) for n={0}, idx={1}".format(n, idx))

if __name__ == '__main__':
    constants = XbarRControlChartConstants()
    A2 = constants.get_constant(4, 0)
    d2 = constants.get_constant(4, 1)
    D3 = constants.get_constant(4, 2)
    D4 = constants.get_constant(4, 3)
    print(A2, d2, D3, D4)



"""
Rule #1 Beyond limits: one or more points are beyond the control limits.
"""

import pandas as pd
from SPC import Rule
pd.options.mode.chained_assignment = None  # default='warn'

class Rule01(Rule):
    rule_id = "R01" # The unique rule idenfitier.

    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """

        # Create a new signal column.
        df[self.rule_id] = False

        # Iterate the rows.
        for i in range(len(df)):
            # Check lower.
            if df['value'][i] < df['LCL'][i]:
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True

            # Check upper.
            if df['value'][i] > df['UCL'][i]:
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True



"""
The abstract class for a rule.
"""

import pandas as pd
from abc import ABC, abstractmethod
 
class Rule(ABC):
    @abstractmethod
    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """
        pass



"""
Rule #3 Zone B: 4 out of 5 consecutive points in Zone B or beyond.
"""

import pandas as pd
from SPC import Rule
pd.options.mode.chained_assignment = None  # default='warn'

class Rule03(Rule):
    rule_id = "R03" # The unique rule idenfitier.

    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """

        # Create a new signal column.
        df[self.rule_id] = False

        # Iterate the rows.
        for i in range(5, len(df)):
            # Check lower.
            if((df['value'][i-4] < df['-1s'][i-4] and df['value'][i-3] < df['-1s'][i-3] and df['value'][i-2] < df['-1s'][i-2] and df['value'][i-1] < df['-1s'][i-1]) or
               (df['value'][i-4] < df['-1s'][i-4] and df['value'][i-3] < df['-1s'][i-3] and df['value'][i-2] < df['-1s'][i-2] and df['value'][i] < df['-1s'][i]) or
               (df['value'][i-4] < df['-1s'][i-4] and df['value'][i-2] < df['-1s'][i-2] and df['value'][i-1] < df['-1s'][i-1] and df['value'][i] < df['-1s'][i]) or
               (df['value'][i-4] < df['-1s'][i-4] and df['value'][i-3] < df['-1s'][i-3] and df['value'][i-1] < df['-1s'][i-1] and df['value'][i] < df['-1s'][i]) or
               (df['value'][i-3] < df['-1s'][i-3] and df['value'][i-2] < df['-1s'][i-2] and df['value'][i-1] < df['-1s'][i-1] and df['value'][i] < df['-1s'][i])):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True

            # Check upper.
            if((df['value'][i-4] > df['+1s'][i-4] and df['value'][i-3] > df['+1s'][i-3] and df['value'][i-2] > df['+1s'][i-2] and df['value'][i-1] > df['+1s'][i-1]) or
               (df['value'][i-4] > df['+1s'][i-4] and df['value'][i-3] > df['+1s'][i-3] and df['value'][i-2] > df['+1s'][i-2] and df['value'][i] > df['+1s'][i]) or
               (df['value'][i-4] > df['+1s'][i-4] and df['value'][i-2] > df['+1s'][i-2] and df['value'][i-1] > df['+1s'][i-1] and df['value'][i] > df['+1s'][i]) or
               (df['value'][i-4] > df['+1s'][i-4] and df['value'][i-3] > df['+1s'][i-3] and df['value'][i-1] > df['+1s'][i-1] and df['value'][i] > df['+1s'][i]) or
               (df['value'][i-3] > df['+1s'][i-3] and df['value'][i-2] > df['+1s'][i-2] and df['value'][i-1] > df['+1s'][i-1] and df['value'][i] > df['+1s'][i])):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True



"""
The control chart constants for the X Bar S control chart.
"""

from SPC import ControlChartConstants

class XbarSControlChartConstants(ControlChartConstants):
    def __init__(self):
        self._constants = {
            2: [2.659, 0.7979, 0, 3.267],
            3: [1.954, 0.8862, 0, 2.568],
            4: [1.628, 0.9213, 0, 2.266],
            5: [1.427, 0.94, 0, 2.089],
            6: [1.287, 0.9515, 0.03, 1.970],
            7: [1.182, 0.9594, 0.118, 1.882],
            8: [1.099, 0.9650, 0.185, 1.815],
            9: [1.032, 0.9693, 0.239, 1.761],
            10: [0.975, 0.9727, 0.284, 1.716],
            15: [0.789, 0.9823, 0.428, 1.572],
            25: [0.606, 0.9896, 0.565, 1.435]
            # Add more values as needed n: [A3, c4, B3, B4]
        }

    def get_constant(self, n, idx):
        """ Returns the control chart constant.

            :param n: The subgroup size.
            :param idx: The index.
        """
        try:
            value= self._constants[n][idx]
            return value
        except KeyError:
            raise ValueError("invalid parameter value(s) for n={0}, idx={1}".format(n, idx))

if __name__ == '__main__':
    constants = XbarSControlChartConstants()
    A3 = constants.get_constant(4, 0)
    c4 = constants.get_constant(4, 1)
    B3 = constants.get_constant(4, 2)
    B4 = constants.get_constant(4, 3)
    print(A3, c4, B3, B4)



"""
The Process Capability chart.
"""

import numpy as np
from SPC import ProcessCapabilityChart

if __name__ == '__main__':
    data = np.array([[23, 25, 24, 26],
                     [22, 26, 24, 25],
                     [28, 28, 22, 23],
                     [25, 25, 26, 36],
                     [22, 22, 25, 26],
                     [26, 24, 23, 22],
                     [29, 24, 24, 24],
                     [26, 25, 25, 22],
                     [22, 25, 24, 24],
                     [25, 22, 26, 24],
                     [24, 24, 24, 23],
                     [24, 25, 26, 23],
                     [22, 28, 22, 26],
                     [23, 24, 25, 26],
                     [24, 25, 29, 24],
                     [24, 22, 28, 26],
                     [24, 25, 25, 25],
                     [22, 24, 25, 26],
                     [26, 25, 22, 24],
                     [26, 22, 24, 25]])

    # Flatten the data.
    data_flat = np.concatenate(data).ravel().tolist()

    # The Process Capability chart.
    capability = ProcessCapabilityChart(data=data_flat, target=23, LSL=21, USL=28)
    capability.plot(bins=8)
    print("Cp={0}".format(capability.Cp))
    print("Cpk={0}".format(capability.Cpk))
    print("samples={0}".format(capability.num_samples))
    print("mean={0}".format(capability.sample_mean))
    print("stdev={0}".format(capability.sample_std))
    print("max={0}".format(capability.sample_max))
    print("min={0}".format(capability.sample_min))
    print("mean={0}".format(capability.sample_median))
    print("pct_below_LSL={0}".format(capability.pct_below_LSL))
    print("pct_above_USL={0}".format(capability.pct_above_USL))
    print("reject_rate={0}".format(capability.reject_rate))
    print("capable={0}".format(capability.capable(target_cpk=1.33)))



"""
Rule #4 Zone C: 7 or more consecutive points on one side of the average (in Zone C or beyond).
"""

import pandas as pd
from SPC import Rule
pd.options.mode.chained_assignment = None  # default='warn'

class Rule04(Rule):
    rule_id = "R04" # The unique rule idenfitier.

    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """

        # Create a new signal column.
        df[self.rule_id] = False

        # Iterate the rows.
        for i in range(7, len(df)):
            # Check lower.
            if(df['value'][i] < df['CL'][i] and
               df['value'][i-1] < df['CL'][i-1] and
               df['value'][i-2] < df['CL'][i-2] and
               df['value'][i-3] < df['CL'][i-3] and
               df['value'][i-4] < df['CL'][i-4] and
               df['value'][i-5] < df['CL'][i-5] and
               df['value'][i-6] < df['CL'][i-6]):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True

            # Check upper.
            if(df['value'][i] > df['CL'][i] and
               df['value'][i-1] > df['CL'][i-1] and
               df['value'][i-2] > df['CL'][i-2] and
               df['value'][i-3] > df['CL'][i-3] and
               df['value'][i-4] > df['CL'][i-4] and
               df['value'][i-5] > df['CL'][i-5] and
               df['value'][i-6] > df['CL'][i-6]):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True



"""
The Multi-vari chart.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from datetime import datetime

class MultiVariChart():
    _dateformat = "%Y-%m-%d" # The date format.
    _dates = [] # The dates.

    def __init__(self, data:list, xlabel:str="", ylabel:str=""):
        """ Initialization.

            :param data: The data.
            :param xlabel: x-as label.
            :param ylabel: y-as label.
        """
        # Remember the parameters.
        self._data=data
        self._xlabel = xlabel
        self._ylabel = ylabel

        # The number of samples.
        self.number_of_sample = len(data)

        # Initialize the arrays of X bar.
        self.value_X = np.zeros(((self.number_of_sample), 1))

        # Set the values.
        for i in range(self.number_of_sample):
            self.value_X[i] = data[i].mean()

    def plot(self):
        """ Create the plot.
        """
        plt.figure(figsize=(15,10))

        # The x-axis can be numeric or datetime.
        if (len(self._dates) == 0):
            x_values_X = list(range(0, len(self.value_X)))
        else:
            format=self._dateformat
            x_values_X = [datetime.strptime(d, format).date() for d in self._dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(self._dateformat))

        # Plot the vertical lines with min/max.
        for i in range(self.number_of_sample):
            plt.vlines(x_values_X[i], self._data[i].min(), self._data[i].max(), color='lightgrey', linewidth=4)

        plt.plot(x_values_X, self.value_X, marker="o", color="blue", label="")
        plt.title("Multi-Vari Chart")

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Set the y-label.
        plt.ylabel( self._ylabel)

        plt.show()

    @property
    def dates(self):
        """ Returns the dates.
        """
        return self._dates
	
    @dates.setter
    def dates(self, dates:list):
        """ Set the dates for the x-axis.

            :param dates: The dates.
        """
        self._dates = dates

    @property
    def dateformat(self) -> str:
        """ Returns the date format.
        """
        return self._dateformat
	
    @dateformat.setter
    def dateformat(self, dateformat:str):
        """ Sets the date format.

            :param dateformat: The date format.
        """
        self._dateformat = dateformat



"""
An individuals and moving range (X-MR) chart is a pair of control charts for processes
with a subgroup size of one. Used to determine if a process is stable and predictable,
it creates a picture of how the system changes over time. The individual (X) chart displays
individual measurements.

https://sixsigmastudyguide.com/xmr-charts/
"""

import numpy as np
from SPC import Rule01
from SPC import Rule02
from SPC import Rule03
from SPC import Rule04
from SPC import Rule05
from SPC import Rule06
from SPC import Rule07
from SPC import Rule08
from SPC import XmRControlChart

if __name__ == '__main__':
    data = np.array([2.92, 2.96, 2.86, 3.04, 3.07, 2.85, 3.00, 2.92, 2.97, 2.97, 3.09, 3.07, 2.99, 3.06, 3.05, 3.02, 3.07, 2.91, 3.07, 3.20])
    dates = ['21-12-21', '22-12-21', '23-12-21', '24-12-21', '25-12-21', '26-12-21', '27-12-21', '28-12-21', '29-12-21','30-12-21', '31-12-21', '01-01-22', '02-01-22', '03-01-22', '05-01-22', '06-01-22', '07-01-22', '08-01-22', '09-01-22', '10-01-22']
    chart = XmRControlChart(data=data, xlabel="x-label", ylabel_top="y-label-top", ylabel_bottom="y-label-bottom")
    normally_distributed=chart.normally_distributed(data=chart.value_X, significance_level=0.05)
    print("normally_distributed={0}".format(normally_distributed))
    #chart.dates = dates
    #chart.dateformat = "%d-%m-%y"
    #stages=chart.stages(data=chart.value_X, max_stages=4)
    #if stages is not None:
    #    chart.split(stages)
    #chart.split([4, 7])
    #chart.limits=True
    chart.append_rules([Rule01(), Rule02(), Rule03(), Rule04(), Rule05(), Rule06(), Rule07(), Rule08()])
    chart.plot()

    df1 = chart.data(0)
    #print(df1[["CL", "UCL", "LCL"]])
    df2 = chart.data(1)
    #print(df2[["CL", "UCL", "LCL"]])
    #print("stable={0}".format(chart.stable()))



"""
The x-bar and R-chart are control charts used to monitor the mean and variation of a
process based on samples taken in a given time.

X-bar chart: The mean or average change in process over time from subgroup values.
The control limits on the X-Bar brings the sample’s mean and center into consideration.

R-chart: The range of the process over the time from subgroups values.
This monitors the spread of the process over the time.

https://sixsigmastudyguide.com/x-bar-r-control-charts/
"""

import numpy as np
from SPC import Rule01
from SPC import Rule02
from SPC import Rule03
from SPC import Rule04
from SPC import Rule05
from SPC import Rule06
from SPC import Rule07
from SPC import Rule08
from SPC import XbarRControlChart

if __name__ == '__main__':
    data = np.array([[23, 25, 24, 26],
                     [22, 26, 24, 25],
                     [28, 28, 22, 23],
                     [25, 25, 26, 36],
                     [22, 22, 25, 26],
                     [26, 24, 23, 22],
                     [29, 24, 24, 24],
                     [26, 25, 25, 22],
                     [22, 25, 24, 24],
                     [25, 22, 26, 24],
                     [24, 24, 24, 23],
                     [24, 25, 26, 23],
                     [22, 28, 22, 26],
                     [23, 24, 25, 26],
                     [24, 25, 29, 24],
                     [24, 22, 28, 26],
                     [24, 25, 25, 25],
                     [22, 24, 25, 26],
                     [26, 25, 22, 24],
                     [26, 22, 24, 25]])
    dates = ['21-12-21', '22-12-21', '23-12-21', '24-12-21', '25-12-21', '26-12-21', '27-12-21', '28-12-21', '29-12-21','30-12-21', '31-12-21', '01-01-22', '02-01-22', '03-01-22', '05-01-22', '06-01-22', '07-01-22', '08-01-22', '09-01-22', '10-01-22']
    chart = XbarRControlChart(data=data, xlabel="x-label", ylabel_top="y-label-top", ylabel_bottom="y-label-bottom")
    normally_distributed=chart.normally_distributed(data=chart.value_X, significance_level=0.05)
    print("normally_distributed={0}".format(normally_distributed))
    chart.dates = dates
    chart.dateformat = "%d-%m-%y"
    #stages=chart.stages(data=chart.value_X, max_stages=4)
    #if stages is not None:
    #    chart.split(stages)
    #chart.split([5, 11])
    #chart.limits=True
    chart.append_rules([Rule01(), Rule02(), Rule03(), Rule04(), Rule05(), Rule06(), Rule07(), Rule08()])
    chart.plot()

    df1 = chart.data(0)
    #print(df1[["CL", "UCL", "LCL"]])
    df2 = chart.data(1)
    #print(df2[["CL", "UCL", "LCL"]])
    #print("stable={0}".format(chart.stable()))



"""
An Individual moving range (I-MR) chart is used when data is continuous and not collected in subgroups.
An I-MR chart provides process variation over time in graphical method.
I-MR chart is basically two separate charts – Individuals (I) chart and Moving Range (MR) chart, the combination of two charts provides the complete picture of process behavior.

I-Chart: Individual chart displays the individual data points and monitors mean and shifts in
the process when the data points collected at regular intervals of time. This chart will help to
identify the common and assignable causes in the process, if any.

MR Chart: While Individual chart monitors the process mean, the Moving Range chart monitors the
process variation when the data points collected at regular intervals of time. In other words the
moving range chart tracks the absolute difference of each measurement to its previous measurement.

https://sixsigmastudyguide.com/i-mr-chart/
"""

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
from datetime import datetime
from SPC import ControlChart

class ImRControlChart(ControlChart):
    _E2 : float # The Control Chart E2 constant.
    _d2 : float # The Control Chart d2 constant.
    _D3 : float # The Control Chart D3 constant.
    _D4 : float # The Control Chart D4 constant.

    def __init__(self, data:list, xlabel:str="", ylabel_top:str="", ylabel_bottom:str=""):
        """ Initialization.

            :param data: values.
            :param xlabel: x-as label.
            :param ylabel_top: top y-as label.
            :param ylabel_bottom: bottom y-as label.
        """
        # Initialization of the base class.
        super().__init__(2) # X chart and mR chart.

        # Remember the parameters.
        self._data = data
        self._xlabel = xlabel
        self._ylabel_top = ylabel_top
        self._ylabel_bottom = ylabel_bottom

        # Remember the Control chart constants size=2 (comparing the current state with previous state).
        self._E2 = 2.66
        self._d2 = 1.128
        self._D3 = 0
        self._D4 = 3.267

        # The number of samples.
        self.number_of_sample = len(data)

        # Set the values array.
        self.value_I = data

        # Initialize the arrays of I.
        self.cl_I = np.zeros(((self.number_of_sample), 1))
        self.ucl_I = np.zeros(((self.number_of_sample), 1))
        self.lcl_I = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_I = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_I = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_I = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_I = np.zeros(((self.number_of_sample), 1))

        # Initialize the arrays of mR.
        self.value_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.cl_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.ucl_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.lcl_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.two_sigma_plus_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.one_sigma_plus_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.two_sigma_min_mR = np.zeros(((self.number_of_sample - 1), 1))
        self.one_sigma_min_mR = np.zeros(((self.number_of_sample - 1), 1))

        # Calculate the difference between the samples.
        for i in range(self.number_of_sample - 1):
            self.value_mR[i]=abs(self.value_I[i+1] - self.value_I[i])

        # Calculate the UCL, CL, LCL of I.
        self.cl_I[:] = self.value_I.mean() 
        self.ucl_I[:] = self.value_I.mean() + self._E2 * self.value_mR.mean()
        self.lcl_I[:] = self.value_I.mean() - self._E2 * self.value_mR.mean()

        # Calculate the one and two sigma of I.
        self.two_sigma_plus_I[:] = self.value_I.mean() + self._E2 * self.value_mR.mean() * 2/3
        self.one_sigma_plus_I[:] = self.value_I.mean() + self._E2 * self.value_mR.mean() * 1/3
        self.one_sigma_min_I[:] = self.value_I.mean() - self._E2 * self.value_mR.mean() * 1/3
        self.two_sigma_min_I[:] = self.value_I.mean() - self._E2 * self.value_mR.mean() * 2/3

        # Calculate the  UCL, CL, LCL of mR.
        self.cl_mR[:] = self.value_mR.mean() 
        self.ucl_mR[:] = self._D4 * self.value_mR.mean()
        self.lcl_mR[:] = self._D3 * self.value_mR.mean()

        # Calculate the one and two sigma of mR.
        self.two_sigma_plus_mR[:] = self.value_mR.mean() + (self._D4 * self.value_mR.mean() - self.value_mR.mean()) * 2/3
        self.one_sigma_plus_mR[:] = self.value_mR.mean() + (self._D4 * self.value_mR.mean() - self.value_mR.mean()) * 1/3
        self.one_sigma_min_mR[:] = self.value_mR.mean() - (self.value_mR.mean() - self._D3 * self.value_mR.mean()) * 1/3
        self.two_sigma_min_mR[:] = self.value_mR.mean() - (self.value_mR.mean() - self._D3 * self.value_mR.mean()) * 2/3

    def plot(self):
        """ Create the plot.
        """
        plt.figure(figsize=(15,10))

        # The 1st vertical plot.
        plt.subplot(2,1,1)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_I = list(range(0, len(self.value_I)))
        else:
            format=super().dateformat
            x_values_I = [datetime.strptime(d, format).date() for d in super().dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # I chart.
        plt.plot(x_values_I, self.value_I, marker="o", color="k", label="I")
        plt.plot(x_values_I, self.ucl_I, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(0)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_I[i], self.value_I[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_I, self.two_sigma_plus_I, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_I, self.one_sigma_plus_I, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_I, self.cl_I, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_I, self.one_sigma_min_I, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_I, self.two_sigma_min_I, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_I, self.lcl_I, color="r", label="LCL")
        plt.title("I Chart")

        # Set the lower and upper limits for the x-axis.
        plt.xlim(min(x_values_I), max(x_values_I))

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel( self._ylabel_top)

        # The 2nd vertical plot.
        plt.subplot(2,1,2)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_mR = list(range(1, len(self.value_I)))
        else:
            format=super().dateformat
            x_values_mR = [datetime.strptime(d, format).date() for d in super().dates][1:]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # mR chart.
        plt.plot(x_values_mR, self.value_mR, marker="o", color="k", label="R")
        plt.plot(x_values_mR, self.ucl_mR, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(1)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_mR[i], self.value_mR[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_mR, self.two_sigma_plus_mR, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_mR, self.one_sigma_plus_mR, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_mR, self.cl_mR, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_mR, self.one_sigma_min_mR, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_mR, self.two_sigma_min_mR, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_mR, self.lcl_mR, color="r", label="LCL")
        plt.title("mR Chart")

        # Set the lower and upper limits for the x-axis.
        plt.xlim(min(x_values_I), max(x_values_I))

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel(self._ylabel_bottom)

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Show the plot.
        plt.show()

    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        # Include the last index.
        if not self.number_of_sample in stages:
            stages.append(self.number_of_sample)

        # Initialize the arrays of I.
        self.cl_I = np.zeros(((self.number_of_sample), 1))
        self.ucl_I = np.zeros(((self.number_of_sample), 1))
        self.lcl_I = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_I = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_I = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_I = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_I = np.zeros(((self.number_of_sample), 1))

        # Initialize the arrays of mR.
        self.value_mR = np.zeros(((self.number_of_sample-1), 1))
        self.cl_mR = np.zeros(((self.number_of_sample-1), 1))
        self.ucl_mR = np.zeros(((self.number_of_sample-1), 1))
        self.lcl_mR = np.zeros(((self.number_of_sample-1), 1))

        # Calculate the difference between the samples.
        for i in range(self.number_of_sample-1):
            self.value_mR[i]=abs(self.value_I[i+1] - self.value_I[i])

        # Make the calculations on each split, instead of the full data.
        start_index_I = 0
        start_index_mR = 0
        for i in stages:
            # Set the end index.
            end_index_I = i
            end_index_mR = i-1

            # Calculate the UCL, CL, LCL of I.
            self.cl_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() 
            self.ucl_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() + self._E2 * self.value_mR[start_index_mR:end_index_mR].mean()
            self.lcl_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() - self._E2 * self.value_mR[start_index_mR:end_index_mR].mean()

            # Calculate the one and two sigma of I.
            self.two_sigma_plus_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() + self._E2 * self.value_mR[start_index_mR:end_index_mR].mean() * 2/3
            self.one_sigma_plus_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() + self._E2 * self.value_mR[start_index_mR:end_index_mR].mean() * 1/3
            self.one_sigma_min_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() - self._E2 * self.value_mR[start_index_mR:end_index_mR].mean() * 1/3
            self.two_sigma_min_I[start_index_I:end_index_I] = self.value_I[start_index_I:end_index_I].mean() - self._E2 * self.value_mR[start_index_mR:end_index_mR].mean() * 2/3

            # Calculate the  UCL, CL, LCL of mR.
            self.cl_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() 
            self.ucl_mR[start_index_mR:end_index_mR] = self._D4*self.value_mR[start_index_mR:end_index_mR].mean()
            self.lcl_mR[start_index_mR:end_index_mR] = self._D3*self.value_mR[start_index_mR:end_index_mR].mean()

            # Calculate the one and two sigma of mR.
            self.two_sigma_plus_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() + (self._D4 * self.value_mR[start_index_mR:end_index_mR].mean() - self.value_mR[start_index_mR:end_index_mR].mean()) * 2/3
            self.one_sigma_plus_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() + (self._D4 * self.value_mR[start_index_mR:end_index_mR].mean() - self.value_mR[start_index_mR:end_index_mR].mean()) * 1/3
            self.one_sigma_min_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() - (self.value_mR[start_index_mR:end_index_mR].mean() - self._D3 * self.value_mR[start_index_mR:end_index_mR].mean()) * 1/3
            self.two_sigma_min_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() - (self.value_mR[start_index_mR:end_index_mR].mean() - self._D3 * self.value_mR[start_index_mR:end_index_mR].mean()) * 2/3

            # Set the start index.
            start_index_I = end_index_I
            start_index_mR = end_index_mR

    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data (0 = X chart, 1 = mR chart)
        """
        if index == 0: # X chart.
            df = pd.DataFrame(np.column_stack([self.value_I, self.ucl_I, self.two_sigma_plus_I, self.one_sigma_plus_I, self.cl_I, self.one_sigma_min_I, self.two_sigma_min_I, self.lcl_I]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates
                df=df.set_index('date')

            return df
        if index == 1: # R chart.
            df = pd.DataFrame(np.column_stack([self.value_mR, self.ucl_mR, self.two_sigma_plus_mR, self.one_sigma_plus_mR, self.cl_mR, self.one_sigma_min_mR, self.two_sigma_min_mR, self.lcl_mR]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])

            self.execute_rules(df)

            return df

        raise ValueError

    def stable(self):
        """ Returns the stable indicator.
        """
        # Execute the rules.
        df = self.data(0)

        if True in df["SIGNAL"].values:
            return False

        # Execute the rules.
        df = self.data(1)

        if True in df["SIGNAL"].values:
            return False

        return True

    def normally_distributed(self, data:list, significance_level:float):
        """ Check if the data follows normal distribution.
            Returns true when the data did not show evidence of non-normality.
            Returns false when the data is not normally distributed.

            :param data: values.
            :param significance_level: significance level.
        """
        return self._normally_distributed(data, significance_level)



"""
The abstract class for a control chart.
"""

import pandas as pd
import ruptures as rpt
import statsmodels.api as sm
import warnings
from SPC import Rule
from abc import ABC, abstractmethod
from kneed import KneeLocator
from scipy.stats import jarque_bera
from scipy.stats import shapiro
warnings.filterwarnings("ignore")

class ControlChart(ABC):
    _limits = False # The limits indicator for +2s, +1s, -1s, -2s.
    _dateformat = "%Y-%m-%d" # The date format.
    _dates = [] # The dates.
    _rules = list() # The rules.

    def __init__(self, number_of_charts : int):
        """ Initialization.

            :param number_of_charts: The number of charts.
        """
        # The number of charts.
        self._number_of_charts = number_of_charts

    @abstractmethod
    def plot(self):
        """ Plot the chart(s).
        """
        pass

    @abstractmethod
    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        pass

    @abstractmethod
    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data.
        """
        pass

    @property
    def limits(self):
        """ Returns the limits indicator.
        """
        return self._limits

    @limits.setter
    def limits(self, limits):
        """ Set the limits indicator.

            :param limits: The limits indicator.
        """
        self._limits = limits

    @property
    def number_of_charts(self):
        """ Returns the number of charts.
        """
        return self._number_of_charts

    @property
    def dates(self):
        """ Returns the dates.
        """
        return self._dates
	
    @dates.setter
    def dates(self, dates:list):
        """ Set the dates for the x-axis.

            :param dates: The dates.
        """
        self._dates = dates

    @property
    def dateformat(self) -> str:
        """ Returns the date format.
        """
        return self._dateformat
	
    @dateformat.setter
    def dateformat(self, dateformat:str):
        """ Sets the date format.

            :param dateformat: The date format.
        """
        self._dateformat = dateformat

    @abstractmethod
    def stable(self):
        """ Returns the stable indicator.
        """
        pass

    def append_rule(self, rule : Rule):
        """ Append a rule.

            :param rule: The rule.
        """
        # Append the rule.
        self._rules.append(rule)

    def append_rules(self, rules : list[Rule]):
        """ Append the rules.

            :param rules: The rules.
        """
        # Set the rules.
        self._rules = rules

    def execute_rules(self, df: pd.DataFrame):
        """ Rules execution.

            :param df: The dataframe.
        """
        # Create the signal column.
        df["SIGNAL"] = False

        for rule in self._rules:
            rule.execute(df)

    def stages(self, data:list, max_stages:int):
        """ Plot the chart(s).

            :param data: values.
            :param max_stages: the maximum possible of stages.
        """
        # Determining errors for possible number of stages.
        sum_errors = []
        for i in range (1, max_stages):
            model = rpt.Dynp(model="l2") # L2=Euclidean Distance.
            model.fit(data)
            stages = model.predict(n_bkps=i-1)
            error=model.cost.sum_of_costs(stages)
            sum_errors.append(error)

        # Calculate optimal number of stages.
        x = range(1, len(sum_errors)+1)
        kn = KneeLocator(x, sum_errors, curve='convex', direction='decreasing')

        # Determine the indexes of the optimal number of stages.
        if kn.knee is not None:
            stages = model.predict(n_bkps=kn.knee)
            return stages
        else:
            return None

    def _normally_distributed(self, data:list, significance_level:float):
        """ Check if the data follows normal distribution.
            Returns true when the data did not show evidence of non-normality.
            Returns false when the data is not normally distributed.

            :param data: values.
            :param significance_level: significance level.
        """
        if (len(data) > 5000):
            stat,p = jarque_bera(data)
        else:
            stat,p = shapiro(data)

        if p > significance_level:
            return True
        else:
            return False

    def _auto_correlated(self, data:list, delay:int, threshold:float):
        """ Check if the data has a correlation with itself for a specific delay.
            Returns true when the data is correlated with itself.
            Returns false when the data is not correlated with itself.

            :param data: values.
            :param delay: delay.
            :param threshold: correlation threshold.
        """
        lags = range(delay)

        # Calculate auto-correlation.
        acorr = sm.tsa.acf(data, nlags = len(lags))

        if abs(acorr[delay]) >= threshold:
            return True
        else:
            return False



"""
Attribute chart: c chart is also known as the control chart for defects (counting of the number of defects). It is generally used to monitor
the number of defects in constant size units.

https://sixsigmastudyguide.com/attribute-chart-c-chart/
"""

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
from datetime import datetime
from SPC import ControlChart

class CControlChart(ControlChart):
    def __init__(self, c:list, n:list, xlabel:str="", ylabel:str=""):
        """ Initialization.

            :param c: number of defects in the samples.
            :param n: sample sizes.
            :param xlabel: x-as label.
            :param ylabel: y-as label.
        """
        # Initialization of the base class.
        super().__init__(1) # C chart.

        # Remember the parameters.
        self.number_of_defects = len(c)
        self._xlabel = xlabel
        self._ylabel = ylabel

        # Remember the number of defects.
        self.c = c

        # Remember the sample sizes.
        self.n = n

        # Initialize the arrays.
        self.cl_c = np.zeros(((self.number_of_defects), 1))
        self.ucl_c = np.zeros(((self.number_of_defects), 1))
        self.lcl_c = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_min = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_min = np.zeros(((self.number_of_defects), 1))

        # Calculate the UCL, CL, LCL.
        self.cl_c[:] = self.c.mean()
        self.ucl_c[:] = self.c.mean() + 3 * np.sqrt(self.c.mean())
        self.lcl_c[:] = self.c.mean() - 3 * np.sqrt(self.c.mean())

        # Calculate the one and two sigma.
        self.two_sigma_plus[:] = self.c.mean() + 3 * np.sqrt(self.c.mean()) * 2/3
        self.one_sigma_plus[:] = self.c.mean() + 3 * np.sqrt(self.c.mean()) * 1/3
        self.one_sigma_min[:] = self.c.mean() - 3 * np.sqrt(self.c.mean()) * 1/3
        self.two_sigma_min[:] = self.c.mean() - 3 * np.sqrt(self.c.mean()) * 2/3

    def plot(self):
        """ Create the plot.
        """
        plt.figure(figsize=(15,5))

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_C = list(range(0, self.number_of_defects))
        else:
            format=super().dateformat
            x_values_C = [datetime.strptime(d, format).date() for d in super().dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # C chart.
        plt.plot(x_values_C, self.c, marker="o",color="k",label="c")
        plt.plot(x_values_C, self.ucl_c,color="r",label="UCL")

        # Retrieve the data.
        df = self.data(0)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_C[i], self.c[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_C, self.two_sigma_plus, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_C, self.one_sigma_plus, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_C, self.cl_c,color="b",label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_C, self.one_sigma_min, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_C, self.two_sigma_min, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_C, self.lcl_c,color="r",label="LCL")
        plt.title("C Control Chart")

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Set the y-label.
        plt.ylabel( self._ylabel)

        plt.show()

    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        # Include the last index.
        if not self.number_of_defects in stages:
            stages.append(self.number_of_defects)

        # Initialize the arrays.
        self.cl_c = np.zeros(((self.number_of_defects), 1))
        self.ucl_c = np.zeros(((self.number_of_defects), 1))
        self.lcl_c = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_min = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_min = np.zeros(((self.number_of_defects), 1))


        # Make the calculations on each split, instead of the full data.
        start_index = 0
        for i in stages:
            # Set the end index.
            end_index = i

            # Calculate the UCL, CL, LCL.
            self.cl_c[start_index:end_index] = self.c[start_index:end_index].mean()
            self.ucl_c[start_index:end_index] = self.c[start_index:end_index].mean() + 3 * np.sqrt(self.c[start_index:end_index].mean())
            self.lcl_c[start_index:end_index] = self.c[start_index:end_index].mean() - 3 * np.sqrt(self.c[start_index:end_index].mean())

            # Calculate the one and two sigma.
            self.two_sigma_plus[start_index:end_index] = self.c[start_index:end_index].mean() + 3 * np.sqrt(self.c[start_index:end_index].mean()) * 2/3
            self.one_sigma_plus[start_index:end_index] = self.c[start_index:end_index].mean() + 3 * np.sqrt(self.c[start_index:end_index].mean()) * 1/3
            self.one_sigma_min[start_index:end_index] = self.c[start_index:end_index].mean() - 3 * np.sqrt(self.c[start_index:end_index].mean()) * 1/3
            self.two_sigma_min[start_index:end_index] = self.c[start_index:end_index].mean() - 3 * np.sqrt(self.c[start_index:end_index].mean()) * 2/3

            # Set the start index.
            start_index = end_index

    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data (0 = C chart)
        """
        if index == 0: # C chart.
            df = pd.DataFrame(np.column_stack([self.c, self.ucl_c, self.two_sigma_plus, self.one_sigma_plus, self.cl_c, self.one_sigma_min, self.two_sigma_min, self.lcl_c]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates
                df=df.set_index('date')

            return df

        raise ValueError

    def stable(self):
        """ Returns the stable indicator.
        """
        # Execute the rules.
        df = self.data(0)

        if True in df["SIGNAL"].values:
            return False

        return True



"""
The abstract class for control chart constants.
"""

from abc import ABC, abstractmethod

class ControlChartConstants(ABC):
    @abstractmethod
    def get_constant(self, n, idx):
        """ Returns the control chart constant.

            :param n: The subgroup size.
            :param idx: The index.
        """
        pass



"""
An individuals and moving range (X-MR) chart is a pair of control charts for processes
with a subgroup size of one. Used to determine if a process is stable and predictable,
it creates a picture of how the system changes over time. The individual (X) chart displays
individual measurements.

https://sixsigmastudyguide.com/xmr-charts/
"""

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from SPC import ControlChart

class XmRControlChart(ControlChart):
    _A2 : float # The Control Chart A2 constant.
    _d2 : float # The Control Chart d2 constant.
    _D3 : float # The Control Chart D3 constant.
    _D4 : float # The Control Chart D4 constant.

    def __init__(self, data:list, xlabel:str="", ylabel_top:str="", ylabel_bottom:str=""):
        """ Initialization.

            :param data: values.
            :param xlabel: x-as label.
            :param ylabel_top: top y-as label.
            :param ylabel_bottom: bottom y-as label.
        """
        # Initialization of the base class.
        super().__init__(2) # X chart and mR chart.

        # Remember the parameters.
        self._data = data
        self._xlabel = xlabel
        self._ylabel_top = ylabel_top
        self._ylabel_bottom = ylabel_bottom

        # Remember the Control chart constants size=2 (comparing the current state with previous state).
        self._A2 = 0
        self._d2 = 1.128
        self._D3 = 0
        self._D4 = 3.267

        # The number of samples.
        self.number_of_sample = len(data)

        # Set the values array.
        self.value_X = data

        # Initialize the arrays of X.
        self.cl_X = np.zeros(((self.number_of_sample), 1))
        self.ucl_X = np.zeros(((self.number_of_sample), 1))
        self.lcl_X = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_X = np.zeros(((self.number_of_sample), 1))

        # Initialize the arrays of mR.
        self.value_mR = np.zeros(((self.number_of_sample -1), 1))
        self.cl_mR = np.zeros(((self.number_of_sample -1), 1))
        self.ucl_mR = np.zeros(((self.number_of_sample -1), 1))
        self.lcl_mR = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_plus_mR = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_plus_mR = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_min_mR = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_min_mR = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the difference between the samples.
        for i in range(self.number_of_sample -1):
            self.value_mR[i]=abs(self.value_X[i+1] - self.value_X[i])

        # Calculate the UCL, CL, LCL of X.
        self.cl_X[:] = self.value_X.mean() 
        self.ucl_X[:] = self.value_X.mean() + 3 * self.value_mR.mean() / self._d2
        self.lcl_X[:] = self.value_X.mean() - 3 * self.value_mR.mean() / self._d2

        # Calculate the one and two sigma of X.
        self.two_sigma_plus_X[:] = self.value_X.mean() + 3 * self.value_mR.mean() / self._d2 * 2/3
        self.one_sigma_plus_X[:] = self.value_X.mean() + 3 * self.value_mR.mean() / self._d2 * 1/3
        self.one_sigma_min_X[:] = self.value_X.mean() - 3 * self.value_mR.mean() / self._d2 * 1/3
        self.two_sigma_min_X[:] = self.value_X.mean() - 3 * self.value_mR.mean() / self._d2 * 2/3

        # Calculate the  UCL, CL, LCL of mR.
        self.cl_mR[:] = self.value_mR.mean() 
        self.ucl_mR[:] = self._D4 * self.value_mR.mean()
        self.lcl_mR[:] = self._D3 * self.value_mR.mean()

        # Calculate the one and two sigma of mR.
        self.two_sigma_plus_mR[:] = self.value_mR.mean() + (self._D4 * self.value_mR.mean() - self.value_mR.mean()) * 2/3
        self.one_sigma_plus_mR[:] = self.value_mR.mean() + (self._D4 * self.value_mR.mean() - self.value_mR.mean()) * 1/3
        self.one_sigma_min_mR[:] = self.value_mR.mean() - (self.value_mR.mean() - self._D3 * self.value_mR.mean()) * 1/3
        self.two_sigma_min_mR[:] = self.value_mR.mean() - (self.value_mR.mean() - self._D3 * self.value_mR.mean()) * 2/3

    def plot(self):
        """ Create the plot.
        """
        plt.figure(figsize=(15,10))

        # The 1st vertical plot.
        plt.subplot(2,1,1)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_X = list(range(0, len(self.value_X)))
        else:
            format=super().dateformat
            x_values_X = [datetime.strptime(d, format).date() for d in super().dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # X chart.
        plt.plot(x_values_X, self.value_X, marker="o", color="k", label="X")
        plt.plot(x_values_X, self.ucl_X, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(0)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_X[i], self.value_X[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_X, self.two_sigma_plus_X, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_X, self.one_sigma_plus_X, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_X, self.cl_X, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_X, self.one_sigma_min_X, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_X, self.two_sigma_min_X, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_X, self.lcl_X, color="r", label="LCL")
        plt.title("X Chart")

        # Set the lower and upper limits for the x-axis.
        plt.xlim(min(x_values_X), max(x_values_X))

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel( self._ylabel_top)

        # The 2nd vertical plot.
        plt.subplot(2,1,2)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_mR = list(range(1, len(self.value_X)))
        else:
            format=super().dateformat
            x_values_mR = [datetime.strptime(d, format).date() for d in super().dates][1:]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # mR chart.
        plt.plot(x_values_mR, self.value_mR, marker="o", color="k", label="R")
        plt.plot(x_values_mR, self.ucl_mR, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(1)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_mR[i], self.value_mR[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_mR, self.two_sigma_plus_mR, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_mR, self.one_sigma_plus_mR, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_mR, self.cl_mR, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_mR, self.one_sigma_min_mR, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_mR, self.two_sigma_min_mR, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_mR, self.lcl_mR, color="r", label="LCL")
        plt.title("mR Chart")

        # Set the lower and upper limits for the x-axis.
        plt.xlim(min(x_values_X), max(x_values_X))

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel(self._ylabel_bottom)

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Show the plot.
        plt.show()

    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        # Include the last index.
        if not self.number_of_sample in stages:
            stages.append(self.number_of_sample)

        # Initialize the arrays of X.
        self.cl_X = np.zeros(((self.number_of_sample), 1))
        self.ucl_X = np.zeros(((self.number_of_sample), 1))
        self.lcl_X = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_X = np.zeros(((self.number_of_sample), 1))

        # Initialize the arrays of mR.
        self.value_mR = np.zeros(((self.number_of_sample -1), 1))
        self.cl_mR = np.zeros(((self.number_of_sample -1), 1))
        self.ucl_mR = np.zeros(((self.number_of_sample -1), 1))
        self.lcl_mR = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_plus_mR = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_plus_mR = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_min_mR = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_min_mR = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the difference between the samples.
        for i in range(self.number_of_sample -1):
            self.value_mR[i]=abs(self.value_X[i+1] - self.value_X[i])

        # Make the calculations on each split, instead of the full data.
        start_index_X = 0
        start_index_mR = 0
        for i in stages:
            # Set the end index.
            end_index_X = i
            end_index_mR = i-1

            # Calculate the UCL, CL, LCL of X.
            self.cl_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() 
            self.ucl_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() + 3 * self.value_mR[start_index_mR:end_index_mR].mean() / self._d2
            self.lcl_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() - 3 * self.value_mR[start_index_mR:end_index_mR].mean() / self._d2

            # Calculate the one and two sigma of X.
            self.two_sigma_plus_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() + 3 * self.value_mR[start_index_mR:end_index_mR].mean() / self._d2 * 2/3
            self.one_sigma_plus_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() + 3 * self.value_mR[start_index_mR:end_index_mR].mean() / self._d2 * 1/3
            self.one_sigma_min_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() - 3 * self.value_mR[start_index_mR:end_index_mR].mean() / self._d2 * 1/3
            self.two_sigma_min_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() - 3 * self.value_mR[start_index_mR:end_index_mR].mean() / self._d2 * 2/3

            # Calculate the  UCL, CL, LCL of mR.
            self.cl_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() 
            self.ucl_mR[start_index_mR:end_index_mR] = self._D4 * self.value_mR[start_index_mR:end_index_mR].mean()
            self.lcl_mR[start_index_mR:end_index_mR] = self._D3 * self.value_mR[start_index_mR:end_index_mR].mean()

            # Calculate the one and two sigma of mR.
            self.two_sigma_plus_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() + (self._D4 * self.value_mR[start_index_mR:end_index_mR].mean() - self.value_mR[start_index_mR:end_index_mR].mean()) * 2/3
            self.one_sigma_plus_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() + (self._D4 * self.value_mR[start_index_mR:end_index_mR].mean() - self.value_mR[start_index_mR:end_index_mR].mean()) * 1/3
            self.one_sigma_min_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() - (self.value_mR[start_index_mR:end_index_mR].mean() - self._D3 * self.value_mR[start_index_mR:end_index_mR].mean()) * 1/3
            self.two_sigma_min_mR[start_index_mR:end_index_mR] = self.value_mR[start_index_mR:end_index_mR].mean() - (self.value_mR[start_index_mR:end_index_mR].mean() - self._D3 * self.value_mR[start_index_mR:end_index_mR].mean()) * 2/3

            # Set the start index.
            start_index_X = end_index_X
            start_index_mR = end_index_mR

    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data (0 = X chart, 1 = mR chart)
        """
        if index == 0: # X chart.
            df = pd.DataFrame(np.column_stack([self.value_X, self.ucl_X, self.two_sigma_plus_X, self.one_sigma_plus_X, self.cl_X, self.one_sigma_min_X, self.two_sigma_min_X, self.lcl_X]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates
                df=df.set_index('date')

            return df
        if index == 1: # mR chart.
            df = pd.DataFrame(np.column_stack([self.value_mR, self.ucl_mR, self.two_sigma_plus_mR, self.one_sigma_plus_mR, self.cl_mR, self.one_sigma_min_mR, self.two_sigma_min_mR, self.lcl_mR]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])

            self.execute_rules(df)

            return df

        raise ValueError

    def stable(self):
        """ Returns the stable indicator.
        """
        # Execute the rules.
        df = self.data(0)

        if True in df["SIGNAL"].values:
            return False

        # Execute the rules.
        df = self.data(1)

        if True in df["SIGNAL"].values:
            return False

        return True

    def normally_distributed(self, data:list, significance_level:float):
        """ Check if the data follows normal distribution.
            Returns true when the data did not show evidence of non-normality.
            Returns false when the data is not normally distributed.

            :param data: values.
            :param significance_level: significance level.
        """
        return self._normally_distributed(data, significance_level)



"""
Rule #2 Zone A: 2 out of 3 consecutive points in Zone A or beyond.
"""

import pandas as pd
from SPC import Rule
pd.options.mode.chained_assignment = None  # default='warn'

class Rule02(Rule):
    rule_id = "R02" # The unique rule idenfitier.

    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """

        # Create a new signal column.
        df[self.rule_id] = False

        # Iterate the rows.
        for i in range(3, len(df)):
            # Check lower.
            if((df['value'][i] < df['-2s'][i] and df['value'][i-1] < df['-2s'][i-1]) or 
               (df['value'][i-1] < df['-2s'][i-1] and df['value'][i-2] < df['-2s'][i-2]) or
               (df['value'][i] < df['-2s'][i] and df['value'][i-2] < df['-2s'][i-2])):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True

            # Check upper.
            if((df['value'][i] > df['+2s'][i] and df['value'][i-1] > df['+2s'][i-1]) or
               (df['value'][i-1] > df['+2s'][i-1] and df['value'][i-2] > df['+2s'][i-2]) or
               (df['value'][i] > df['+2s'][i] and df['value'][i-2] > df['+2s'][i-2])):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True



"""

"""

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from SPC import ControlChart
from SPC import XbarRControlChartConstants

class XbarmRControlChart(ControlChart):
    _A2 : float # The Control Chart A2 constant.
    _d2 : float # The Control Chart d2 constant.
    _D3 : float # The Control Chart D3 constant.
    _D4 : float # The Control Chart D4 constant.

    def __init__(self, data:list, xlabel:str="", ylabel_top:str="", ylabel_bottom:str=""):
        """ Initialization.

            :param data: values.
            :param xlabel: x-as label.
            :param ylabel_top: top y-as label.
            :param ylabel_bottom: bottom y-as label.
        """
        # Initialization of the base class.
        super().__init__(2) # X bar chart and mR chart.

        # Remember the parameters.
        self._data = data
        self._xlabel = xlabel
        self._ylabel_top = ylabel_top
        self._ylabel_bottom = ylabel_bottom

        # Determine the subgroup size.
        n = len(data[0])

        # Set the values array.
        self.value_X = data

        # The number of samples.
        self.number_of_sample = len(data)

        # Initialize the sample averages.
        self._sample_average = np.zeros(((self.number_of_sample), 1))

        # Calculate the sample averages.
        for i in range(self.number_of_sample):
            self._sample_average[i]=np.average(data[i])

        # Initialize the moving range average.
        self._value_mR_average = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the moving range average.
        for i in range(self.number_of_sample -1):
            self._value_mR_average[i]=abs(self._sample_average[i+1] - self._sample_average[i])

        # Initialize the range.
        self._range = np.zeros(((self.number_of_sample), 1))

        # Calculate the range.
        for i in range(self.number_of_sample):
            self._range[i]=np.max(data[i]) - np.min(data[i])

        # Initialize the moving range.
        self._value_mR = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the moving range average.
        for i in range(self.number_of_sample -1):
            self._value_mR[i]=abs(self._range[i+1] - self._range[i])

        # Initialize the arrays of X bar (MR method).
        self.cl_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.ucl_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.lcl_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_plus_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_plus_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_min_Xmr = np.zeros(((self.number_of_sample) -1, 1))
        self.one_sigma_min_Xmr = np.zeros(((self.number_of_sample -1), 1))

        # Initialize the arrays of R (MR method).
        self.cl_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.ucl_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.lcl_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_plus_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_plus_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_min_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_min_Rmr = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the UCL, CL, LCL of X bar (MR method).
        self.cl_Xmr[:] = self._sample_average.mean()
        self.ucl_Xmr[:] = self._sample_average.mean() + 2.66 * self._value_mR_average.mean()
        self.lcl_Xmr[:] = max(0, self._sample_average.mean() - 2.66 * self._value_mR_average.mean())

        # Calculate the one and two sigma of X bar (MR method).
        self.two_sigma_plus_Xmr[:] = self._sample_average.mean() + 2.66 * self._value_mR_average.mean() * 2/3
        self.one_sigma_plus_Xmr[:] = self._sample_average.mean() + 2.66 * self._value_mR_average.mean() * 1/3
        self.one_sigma_min_Xmr[:] = max(0, self._sample_average.mean() - 2.66 * self._value_mR_average.mean() * 1/3)
        self.two_sigma_min_Xmr[:] = max(0, self._sample_average.mean() - 2.66 * self._value_mR_average.mean() * 2/3)

        # Calculate the UCL, CL, LCL of R (MR method).
        self.cl_Rmr[:] = self._range.mean() 
        self.ucl_Rmr[:] = self._range.mean() + 2.66 * self._value_mR.mean()
        self.lcl_Rmr[:] = max(0, self._range.mean() - 2.66 * self._value_mR.mean())

        # Calculate the one and two sigma of R (MR method).
        self.two_sigma_plus_Rmr[:] = self._range.mean() + 2.66 * self._value_mR.mean() * 2/3
        self.one_sigma_plus_Rmr[:] = self._range.mean() + 2.66 * self._value_mR.mean() * 1/3
        self.one_sigma_min_Rmr[:] = max(0, self._range.mean() - 2.66 * self._value_mR.mean() * 1/3)
        self.two_sigma_min_Rmr[:] = max(0, self._range.mean() - 2.66 * self._value_mR.mean() * 2/3)

    def plot(self):
        """ Create the plot.
        """
        plt.figure(figsize=(15,10))

        # The 1st vertical plot.
        plt.subplot(2,1,1)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_X = list(range(0, len(self._value_mR_average)))
        else:
            format=super().dateformat
            x_values_X = [datetime.strptime(d, format).date() for d in super().dates][1:]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # X chart.
        plt.plot(x_values_X, self._value_mR_average, marker="o", color="k", label="X")
        plt.plot(x_values_X, self.ucl_Xmr, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(0)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_X[i], self._value_mR_average[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_X, self.two_sigma_plus_Xmr, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_X, self.one_sigma_plus_Xmr, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_X, self.cl_Xmr, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_X, self.one_sigma_min_Xmr, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_X, self.two_sigma_min_Xmr, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_X, self.lcl_Xmr, color="r", label="LCL")
        plt.title("X (MR) Chart")

        # Set the lower and upper limits for the x-axis.
        plt.xlim(min(x_values_X), max(x_values_X))

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel( self._ylabel_top)

        # The 2nd vertical plot.
        plt.subplot(2,1,2)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_mR = list(range(0, len(self._value_mR)))
        else:
            format=super().dateformat
            x_values_mR = [datetime.strptime(d, format).date() for d in super().dates][1:]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # R (MR method) chart.
        plt.plot(x_values_mR, self._value_mR, marker="o", color="k", label="R")
        plt.plot(x_values_mR, self.ucl_Rmr, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(1)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_mR[i], self._value_mR[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_mR, self.two_sigma_plus_Rmr, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_mR, self.one_sigma_plus_Rmr, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_mR, self.cl_Rmr, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_mR, self.one_sigma_min_Rmr, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_mR, self.two_sigma_min_Rmr, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_mR, self.lcl_Rmr, color="r", label="LCL")
        plt.title("R (MR) Chart")

        # Set the lower and upper limits for the x-axis.
        plt.xlim(min(x_values_X), max(x_values_X))

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel(self._ylabel_bottom)

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Show the plot.
        plt.show()

    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        # Include the last index.
        if not self.number_of_sample in stages:
            stages.append(self.number_of_sample)

        # Initialize the arrays of X bar (MR method).
        self.cl_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.ucl_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.lcl_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_plus_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_plus_Xmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_min_Xmr = np.zeros(((self.number_of_sample) -1, 1))
        self.one_sigma_min_Xmr = np.zeros(((self.number_of_sample -1), 1))

        # Initialize the arrays of R (MR method).
        self.cl_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.ucl_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.lcl_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_plus_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_plus_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.two_sigma_min_Rmr = np.zeros(((self.number_of_sample -1), 1))
        self.one_sigma_min_Rmr = np.zeros(((self.number_of_sample -1), 1))

        # Initialize the sample averages.
        self._sample_average = np.zeros(((self.number_of_sample), 1))

        # Calculate the sample averages.
        for i in range(self.number_of_sample):
            self._sample_average[i]=np.average(self.value_X[i])

        # Initialize the moving range average.
        self._value_mR_average = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the moving range average.
        for i in range(self.number_of_sample -1):
            self._value_mR_average[i]=abs(self._sample_average[i+1] - self._sample_average[i])

        # Initialize the range.
        self._range = np.zeros(((self.number_of_sample), 1))

        # Calculate the range.
        for i in range(self.number_of_sample):
            self._range[i]=np.max(self.value_X[i]) - np.min(self.value_X[i])

        # Initialize the moving range.
        self._value_mR = np.zeros(((self.number_of_sample -1), 1))

        # Calculate the moving range average.
        for i in range(self.number_of_sample -1):
            self._value_mR[i]=abs(self._range[i+1] - self._range[i])

        # Make the calculations on each split, instead of the full data.
        start_index_mR = 0
        start_index_Rmr = 0
        for i in stages:
            # Set the end index.
            end_index_mR = i
            end_index_Rmr = i

            print(f'start_index_mR={start_index_mR}, end_index_mR={end_index_mR}')
            print(f'start_index_Rmr={start_index_Rmr}, end_index_Rmr={end_index_Rmr}')

            # Calculate the UCL, CL, LCL of X bar (MR method).
            self.cl_Xmr[start_index_mR:end_index_mR] = self._sample_average[start_index_mR:end_index_mR].mean()
            self.ucl_Xmr[start_index_mR:end_index_mR] = self._sample_average[start_index_mR:end_index_mR].mean() + 2.66 * self._value_mR_average[start_index_mR:end_index_mR].mean()
            self.lcl_Xmr[start_index_mR:end_index_mR] = max(0, self._sample_average[start_index_mR:end_index_mR].mean() - 2.66 * self._value_mR_average[start_index_mR:end_index_mR].mean())

            # Calculate the one and two sigma of X bar (MR method).
            self.two_sigma_plus_Xmr[start_index_mR:end_index_mR] = self._sample_average[start_index_mR:end_index_mR].mean() + 2.66 * self._value_mR_average[start_index_mR:end_index_mR].mean() * 2/3
            self.one_sigma_plus_Xmr[start_index_mR:end_index_mR] = self._sample_average[start_index_mR:end_index_mR].mean() + 2.66 * self._value_mR_average[start_index_mR:end_index_mR].mean() * 1/3
            self.one_sigma_min_Xmr[start_index_mR:end_index_mR] = max(0, self._sample_average[start_index_mR:end_index_mR].mean() - 2.66 * self._value_mR_average[start_index_mR:end_index_mR].mean() * 1/3)
            self.two_sigma_min_Xmr[start_index_mR:end_index_mR] = max(0, self._sample_average[start_index_mR:end_index_mR].mean() - 2.66 * self._value_mR_average[start_index_mR:end_index_mR].mean() * 2/3)

            # Calculate the UCL, CL, LCL of R (MR method).
            self.cl_Rmr[start_index_Rmr:end_index_Rmr] = self._range[start_index_Rmr:end_index_Rmr].mean() 
            self.ucl_Rmr[start_index_Rmr:end_index_Rmr] = self._range[start_index_Rmr:end_index_Rmr].mean() + 2.66 * self._value_mR[start_index_Rmr:end_index_Rmr].mean()
            self.lcl_Rmr[start_index_Rmr:end_index_Rmr] = max(0, self._range[start_index_Rmr:end_index_Rmr].mean() - 2.66 * self._value_mR[start_index_Rmr:end_index_Rmr].mean())

            # Calculate the one and two sigma of R (MR method).
            self.two_sigma_plus_Rmr[start_index_Rmr:end_index_Rmr] = self._range[start_index_Rmr:end_index_Rmr].mean() + 2.66 * self._value_mR[start_index_Rmr:end_index_Rmr].mean() * 2/3
            self.one_sigma_plus_Rmr[start_index_Rmr:end_index_Rmr] = self._range[start_index_Rmr:end_index_Rmr].mean() + 2.66 * self._value_mR[start_index_Rmr:end_index_Rmr].mean() * 1/3
            self.one_sigma_min_Rmr[start_index_Rmr:end_index_Rmr] = max(0, self._range[start_index_Rmr:end_index_Rmr].mean() - 2.66 * self._value_mR[start_index_Rmr:end_index_Rmr].mean() * 1/3)
            self.two_sigma_min_Rmr[start_index_Rmr:end_index_Rmr] = max(0, self._range[start_index_Rmr:end_index_Rmr].mean() - 2.66 * self._value_mR[start_index_Rmr:end_index_Rmr].mean() * 2/3)

            # Set the start index.
            start_index_mR = end_index_mR
            start_index_Rmr = end_index_Rmr

    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data (0 = X chart, 1 = mR chart)
        """
        if index == 0: # X chart.
            df = pd.DataFrame(np.column_stack([self._value_mR_average, self.ucl_Xmr, self.two_sigma_plus_Xmr, self.one_sigma_plus_Xmr, self.cl_Xmr, self.one_sigma_min_Xmr, self.two_sigma_min_Xmr, self.lcl_Xmr]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates[1:]
                df=df.set_index('date')

            return df
        if index == 1: # mR chart.
            df = pd.DataFrame(np.column_stack([self._value_mR, self.ucl_Rmr, self.two_sigma_plus_Rmr, self.one_sigma_plus_Rmr, self.cl_Rmr, self.one_sigma_min_Rmr, self.two_sigma_min_Rmr, self.lcl_Rmr]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates[1:]
                df=df.set_index('date')

            return df

        raise ValueError

    def stable(self):
        """ Returns the stable indicator.
        """
        # Execute the rules.
        df = self.data(0)

        if True in df["SIGNAL"].values:
            return False

        # Execute the rules.
        df = self.data(1)

        if True in df["SIGNAL"].values:
            return False

        return True

    def normally_distributed(self, significance_level:float):
        """ Check if the data follows normal distribution.
            Returns true when the data did not show evidence of non-normality.
            Returns false when the data is not normally distributed.

            :param significance_level: significance level.
        """
        return self._normally_distributed(self._sample_average, significance_level)

    def auto_correlated(self, delay:int, threshold:float):
        """ Check if the data has a correlation with itself for a specific delay.
            Returns true when the data is correlated with itself.
            Returns false when the data is not correlated with itself.

            :param delay: delay.
            :param threshold: correlation threshold.
        """
        return self._auto_correlated(self._sample_average, delay, threshold)



# This will allow shorter imports.
from SPC.Rule import Rule
from SPC.ControlChart import ControlChart
from SPC.ControlChartConstants import ControlChartConstants
from SPC.XbarRControlChartConstants import XbarRControlChartConstants
from SPC.XbarSControlChartConstants import XbarSControlChartConstants
from SPC.MultiVariChart import MultiVariChart
from SPC.ProcessCapabilityChart import ProcessCapabilityChart
from SPC.Rule01 import Rule01
from SPC.Rule02 import Rule02
from SPC.Rule03 import Rule03
from SPC.Rule04 import Rule04
from SPC.Rule05 import Rule05
from SPC.Rule06 import Rule06
from SPC.Rule07 import Rule07
from SPC.Rule08 import Rule08
from SPC.XbarmRControlChart import XbarmRControlChart
from SPC.XbarRControlChart import XbarRControlChart
from SPC.XbarSControlChart import XbarSControlChart
from SPC.XmRControlChart import XmRControlChart
from SPC.ImRControlChart import ImRControlChart
from SPC.UControlChart import UControlChart
from SPC.CControlChart import CControlChart
from SPC.PControlChart import PControlChart
from SPC.NPControlChart import NPControlChart



"""
Rule #6 Mixture: 8 consecutive points with no points in Zone C.
"""

import pandas as pd
from SPC import Rule
pd.options.mode.chained_assignment = None  # default='warn'

class Rule06(Rule):
    rule_id = "R06" # The unique rule idenfitier.

    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """

        # Create a new signal column.
        df[self.rule_id] = False

        # Iterate the rows.
        for i in range(8, len(df)):
            # Check.
            if((df['value'][i] < df['-1s'][i] or df['value'][i] > df['+1s'][i]) and
               (df['value'][i-1] < df['-1s'][i-1] or df['value'][i-1] > df['+1s'][i-1]) and
               (df['value'][i-2] < df['-1s'][i-2] or df['value'][i-2] > df['+1s'][i-2]) and
               (df['value'][i-3] < df['-1s'][i-3] or df['value'][i-3] > df['+1s'][i-3]) and
               (df['value'][i-4] < df['-1s'][i-4] or df['value'][i-4] > df['+1s'][i-4]) and
               (df['value'][i-5] < df['-1s'][i-5] or df['value'][i-5] > df['+1s'][i-5]) and
               (df['value'][i-6] < df['-1s'][i-6] or df['value'][i-6] > df['+1s'][i-6]) and
               (df['value'][i-7] < df['-1s'][i-7] or df['value'][i-7] > df['+1s'][i-7])):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True



"""
An Individual moving range (I-MR) chart is used when data is continuous and not collected in subgroups.
An I-MR chart provides process variation over time in graphical method.
I-MR chart is basically two separate charts – Individuals (I) chart and Moving Range (MR) chart, the combination of two charts provides the complete picture of process behavior.

I-Chart: Individual chart displays the individual data points and monitors mean and shifts in
the process when the data points collected at regular intervals of time. This chart will help to
identify the common and assignable causes in the process, if any.

MR Chart: While Individual chart monitors the process mean, the Moving Range chart monitors the
process variation when the data points collected at regular intervals of time. In other words the
moving range chart tracks the absolute difference of each measurement to its previous measurement.

https://sixsigmastudyguide.com/i-mr-chart/
"""

import numpy as np
from SPC import Rule01
from SPC import Rule02
from SPC import Rule03
from SPC import Rule04
from SPC import Rule05
from SPC import Rule06
from SPC import Rule07
from SPC import Rule08
from SPC import ImRControlChart

if __name__ == '__main__':
    data = np.array([82, 84, 75, 79, 84, 81, 81, 82, 80, 78, 74])
    dates = ['21-12-21', '22-12-21', '23-12-21', '24-12-21', '25-12-21', '26-12-21', '27-12-21', '28-12-21', '29-12-21','30-12-21', '31-12-21']
    chart = ImRControlChart(data=data, xlabel="x-label", ylabel_top="y-label-top", ylabel_bottom="y-label-bottom")
    normally_distributed=chart.normally_distributed(data=chart.value_I, significance_level=0.05)
    print("normally_distributed={0}".format(normally_distributed))
    #chart.dates = dates
    #chart.dateformat = "%d-%m-%y"
    #stages=chart.stages(data=chart.value_I, max_stages=2)
    #if stages is not None:
    #    chart.split(stages)
    #chart.split([4, 7])
    #chart.limits=True
    chart.append_rules([Rule01(), Rule02(), Rule03(), Rule04(), Rule05(), Rule06(), Rule07(), Rule08()])
    chart.plot()

    df1 = chart.data(0)
    #print(df1[["CL", "UCL", "LCL"]])
    df2 = chart.data(1)
    #print(df2[["CL", "UCL", "LCL"]])
    #print("stable={0}".format(chart.stable()))



"""
Rule #8 Over-control: 14 consecutive points alternating up and down.
"""

import pandas as pd
from SPC import Rule
pd.options.mode.chained_assignment = None  # default='warn'

class Rule08(Rule):
    rule_id = "R08" # The unique rule idenfitier.

    def execute(self, df: pd.DataFrame):
        """ Rule execution.

            :param df: The dataframe.
        """
        # Create a new signal column.
        df[self.rule_id] = False

        # Iterate the rows.
        for i in range(14, len(df)):
            # Check.
            if(((df['value'][i] > df['value'][i-1]) and
               (df['value'][i-1] < df['value'][i-2]) and
               (df['value'][i-2] > df['value'][i-3]) and
               (df['value'][i-3] < df['value'][i-4]) and
               (df['value'][i-4] > df['value'][i-5]) and
               (df['value'][i-5] < df['value'][i-6]) and
               (df['value'][i-6] > df['value'][i-7]) and
               (df['value'][i-7] < df['value'][i-8]) and
               (df['value'][i-8] > df['value'][i-9]) and
               (df['value'][i-9] < df['value'][i-10]) and
               (df['value'][i-10] > df['value'][i-11]) and
               (df['value'][i-11] < df['value'][i-12]) and
               (df['value'][i-12] > df['value'][i-13])) or
               ((df['value'][i] < df['value'][i-1]) and
               (df['value'][i-1] > df['value'][i-2]) and
               (df['value'][i-2] < df['value'][i-3]) and
               (df['value'][i-3] > df['value'][i-4]) and
               (df['value'][i-4] < df['value'][i-5]) and
               (df['value'][i-5] > df['value'][i-6]) and
               (df['value'][i-6] < df['value'][i-7]) and
               (df['value'][i-7] > df['value'][i-8]) and
               (df['value'][i-8] < df['value'][i-9]) and
               (df['value'][i-9] > df['value'][i-10]) and
               (df['value'][i-10] < df['value'][i-11]) and
               (df['value'][i-11] > df['value'][i-12]) and
               (df['value'][i-12] < df['value'][i-13]))):
                df[self.rule_id][i] = True
                df["SIGNAL"][i] = True



"""
X Bar S charts are control charts to examine the process mean and standard deviation over the time.

X-bar chart: The mean or average change in process over time from subgroup values.
The control limits on the X-Bar brings the sample’s mean and center into consideration.

S-chart: The standard deviation of the process over the time from subgroups values.
This monitors the process standard deviation (as approximated by the sample moving range).

https://sixsigmastudyguide.com/x-bar-s-chart/
"""

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
from datetime import datetime
from SPC import ControlChart
from SPC import XbarSControlChartConstants

class XbarSControlChart(ControlChart):
    _A3 : float # The Control Chart A3 constant.
    _c4 : float # The Control Chart c4 constant.
    _B3 : float # The Control Chart B3 constant.
    _B4 : float # The Control Chart B4 constant.

    def __init__(self, data:list, xlabel:str="", ylabel_top:str="", ylabel_bottom:str=""):
        """ Initialization.

            :param data: values.
            :param xlabel: x-as label.
            :param ylabel_top: top y-as label.
            :param ylabel_bottom: bottom y-as label.
        """
        # Initialization of the base class.
        super().__init__(2) # X bar chart and S chart.

        # Remember the parameters.
        self._data = data
        self._xlabel = xlabel
        self._ylabel_top = ylabel_top
        self._ylabel_bottom = ylabel_bottom

        # Initialization of the constants.
        constants = XbarSControlChartConstants()

        # Determine the subgroup size.
        n = len(data[0])

        # Remember the Control chart constants.
        self._A3 = constants.get_constant(n, 0)
        self._c4 = constants.get_constant(n, 1)
        self._B3 = constants.get_constant(n, 2)
        self._B4 = constants.get_constant(n, 3)

        # The number of samples.
        self.number_of_sample = len(data)

        # Initialize the arrays of X bar.
        self.value_X = np.zeros(((self.number_of_sample), 1))
        self.cl_Xbar = np.zeros(((self.number_of_sample), 1))
        self.ucl_Xbar = np.zeros(((self.number_of_sample), 1))
        self.lcl_Xbar = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_X = np.zeros(((self.number_of_sample), 1))

        # Initialize the arrays of S.
        self.value_S = np.zeros(((self.number_of_sample), 1))
        self.cl_S = np.zeros(((self.number_of_sample), 1))
        self.ucl_S = np.zeros(((self.number_of_sample), 1))
        self.lcl_S = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_S = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_S = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_S = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_S = np.zeros(((self.number_of_sample), 1))

        # Set the values for X bar and S.
        for i in range(self.number_of_sample):
            self.value_X[i] = data[i].mean()
            self.value_S[i] = data[i].std(ddof=1) # sample standard deviation.

        # Calculate the UCL, CL, LCL of X bar.
        self.cl_Xbar[:] = self.value_X.mean() 
        self.ucl_Xbar[:] = self.value_X.mean() + self._A3 * self.value_S.mean()
        self.lcl_Xbar[:] = self.value_X.mean() - self._A3 * self.value_S.mean()

        # Calculate the one and two sigma of X bar.
        self.two_sigma_plus_X[:] = self.value_X.mean() + self._A3 * self.value_S.mean() * 2/3
        self.one_sigma_plus_X[:] = self.value_X.mean() + self._A3 * self.value_S.mean() * 1/3
        self.one_sigma_min_X[:] = self.value_X.mean() - self._A3 * self.value_S.mean() * 1/3
        self.two_sigma_min_X[:] = self.value_X.mean() - self._A3 * self.value_S.mean() * 2/3

        # Calculate the  UCL, CL, LCL of S.
        self.cl_S[:] = self.value_S.mean() 
        self.ucl_S[:] = self._B4 * self.value_S.mean()
        self.lcl_S[:] = self._B3 * self.value_S.mean()

        # Calculate the one and two sigma of S.
        self.two_sigma_plus_S[:] = self.value_S.mean() + (self._B4 * self.value_S.mean() - self.value_S.mean()) * 2/3
        self.one_sigma_plus_S[:] = self.value_S.mean() + (self._B4 * self.value_S.mean() - self.value_S.mean()) * 1/3
        self.one_sigma_min_S[:] = self.value_S.mean() - (self.value_S.mean() - self._B3 * self.value_S.mean()) * 1/3
        self.two_sigma_min_S[:] = self.value_S.mean() - (self.value_S.mean() - self._B3 * self.value_S.mean()) * 2/3

    def plot(self):
        """ Create the plot.
        """
        plt.figure(figsize=(15,10))

        # The 1st vertical plot.
        plt.subplot(2,1,1)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_X = list(range(0, len(self.value_X)))
        else:
            format=super().dateformat
            x_values_X = [datetime.strptime(d, format).date() for d in super().dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # X bar chart.
        plt.plot(x_values_X, self.value_X, marker="o", color="k", label="X bar")
        plt.plot(x_values_X, self.ucl_Xbar, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(0)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_X[i], self.value_X[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_X, self.two_sigma_plus_X, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_X, self.one_sigma_plus_X, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_X, self.cl_Xbar, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_X, self.one_sigma_min_X, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_X, self.two_sigma_min_X, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_X, self.lcl_Xbar, color="r", label="LCL")
        plt.title("X bar Chart")

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel( self._ylabel_top)

        # The 2nd vertical plot.
        plt.subplot(2,1,2)

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_S = list(range(0, len(self.value_S)))
        else:
            format=super().dateformat
            x_values_S = [datetime.strptime(d, format).date() for d in super().dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # S chart.
        plt.plot(x_values_S, self.value_S, marker="o", color="k", label="S")
        plt.plot(x_values_S, self.ucl_S, color="r", label="UCL")

        # Retrieve the data.
        df = self.data(1)

        # Plot the signals.
        for i in np.where(df["SIGNAL"])[0]:
            plt.plot(x_values_S[i], self.value_S[i], marker="s", color="r")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_S, self.two_sigma_plus_S, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_S, self.one_sigma_plus_S, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_S, self.cl_S, color="b", label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_S, self.one_sigma_min_S, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_S, self.two_sigma_min_S, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_S, self.lcl_S, color="r", label="LCL")
        plt.title("S Chart")

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the y-label.
        plt.ylabel(self._ylabel_bottom)

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Show the plot.
        plt.show()

    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        # Include the last index.
        if not self.number_of_sample in stages:
            stages.append(self.number_of_sample)

        # Initialize the arrays of X bar.
        self.cl_Xbar = np.zeros(((self.number_of_sample), 1))
        self.ucl_Xbar = np.zeros(((self.number_of_sample), 1))
        self.lcl_Xbar = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_X = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_X = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_X = np.zeros(((self.number_of_sample), 1))

        # Initialize the arrays of S.
        self.value_S = np.zeros(((self.number_of_sample), 1))
        self.cl_S = np.zeros(((self.number_of_sample), 1))
        self.ucl_S = np.zeros(((self.number_of_sample), 1))
        self.lcl_S = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_plus_S = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_plus_S = np.zeros(((self.number_of_sample), 1))
        self.two_sigma_min_S = np.zeros(((self.number_of_sample), 1))
        self.one_sigma_min_S = np.zeros(((self.number_of_sample), 1))

        # Set the values for X bar and S.
        for i in range(self.number_of_sample):
            self.value_X[i] = self._data[i].mean()
            self.value_S[i] = self._data[i].std(ddof=1) # sample standard deviation.

        # Make the calculations on each split, instead of the full data.
        start_index_X = 0
        start_index_S = 0
        for i in stages:
            # Set the end index.
            end_index_X = i
            end_index_S = i

            # Calculate the UCL, CL, LCL of X.
            self.cl_Xbar[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean()
            self.ucl_Xbar[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() + self._A3 * self.value_S[start_index_S:end_index_S].mean()
            self.lcl_Xbar[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() - self._A3 * self.value_S[start_index_S:end_index_S].mean()

            # Calculate the one and two sigma of X.
            self.two_sigma_plus_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() + self._A3 * self.value_S[start_index_S:end_index_S].mean() * 2/3
            self.one_sigma_plus_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() + self._A3 * self.value_S[start_index_S:end_index_S].mean() * 1/3
            self.one_sigma_min_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() - self._A3 * self.value_S[start_index_S:end_index_S].mean() * 1/3
            self.two_sigma_min_X[start_index_X:end_index_X] = self.value_X[start_index_X:end_index_X].mean() - self._A3 * self.value_S[start_index_S:end_index_S].mean() * 2/3

            # Calculate the  UCL, CL, LCL of S.
            self.cl_S[start_index_S:end_index_S] = self.value_S[start_index_S:end_index_S].mean() 
            self.ucl_S[start_index_S:end_index_S] = self._B4 * self.value_S[start_index_S:end_index_S].mean()
            self.lcl_S[start_index_S:end_index_S] = self._B3 * self.value_S[start_index_S:end_index_S].mean()

            # Calculate the one and two sigma of S.
            self.two_sigma_plus_S[start_index_S:end_index_S] = self.value_S[start_index_S:end_index_S].mean() + (self._B4*self.value_S[start_index_S:end_index_S].mean() - self.value_S[start_index_S:end_index_S].mean()) * 2/3
            self.one_sigma_plus_S[start_index_S:end_index_S] = self.value_S[start_index_S:end_index_S].mean() + (self._B4*self.value_S[start_index_S:end_index_S].mean() - self.value_S[start_index_S:end_index_S].mean()) * 1/3
            self.one_sigma_min_S[start_index_X:end_index_X] = self.value_S[start_index_S:end_index_S].mean() - (self.value_S[start_index_S:end_index_S].mean() - self._B3*self.value_S[start_index_S:end_index_S].mean()) * 1/3
            self.two_sigma_min_S[start_index_X:end_index_X] = self.value_S[start_index_S:end_index_S].mean() - (self.value_S[start_index_S:end_index_S].mean() - self._B3*self.value_S[start_index_S:end_index_S].mean()) * 2/3

            # Set the start index.
            start_index_X = end_index_X
            start_index_S = end_index_S

    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data (0 = X chart, 1 = mR chart)
        """
        if index == 0: # X chart.
            df = pd.DataFrame(np.column_stack([self.value_X, self.ucl_Xbar, self.two_sigma_plus_X, self.one_sigma_plus_X, self.cl_Xbar, self.one_sigma_min_X, self.two_sigma_min_X, self.lcl_Xbar]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates
                df=df.set_index('date')

            return df
        if index == 1: # R chart.
            df = pd.DataFrame(np.column_stack([self.value_S, self.ucl_S, self.two_sigma_plus_S, self.one_sigma_plus_S, self.cl_S, self.one_sigma_min_S, self.two_sigma_min_S, self.lcl_S]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])

            self.execute_rules(df)

            return df

        raise ValueError

    def stable(self):
        """ Returns the stable indicator.
        """
        # Execute the rules.
        df = self.data(0)

        if True in df["SIGNAL"].values:
            return False

        # Execute the rules.
        df = self.data(1)

        if True in df["SIGNAL"].values:
            return False

        return True

    def normally_distributed(self, data:list, significance_level:float):
        """ Check if the data follows normal distribution.
            Returns true when the data did not show evidence of non-normality.
            Returns false when the data is not normally distributed.

            :param data: values.
            :param significance_level: significance level.
        """
        return self._normally_distributed(data, significance_level)



"""
Attribute charts: p chart is also known as the control chart for proportions.

https://sixsigmastudyguide.com/p-attribute-charts/
"""

import numpy as np
from SPC import Rule01
from SPC import Rule02
from SPC import Rule03
from SPC import Rule04
from SPC import Rule05
from SPC import Rule06
from SPC import Rule07
from SPC import Rule08
from SPC import PControlChart

if __name__ == '__main__':
    c = np.array([18,15,13,16,8,6,18,14,22,12,13,19,33,20,20,20,22,17,19,21])
    n = np.array([1250,1300,1350,1200,1050,1050,1200,1100,1000,600,1350,1250,1200,1100,1050,950,1560,1150,1230,1100])
    dates = ['21-12-21', '22-12-21', '23-12-21', '24-12-21', '25-12-21', '26-12-21', '27-12-21', '28-12-21', '29-12-21','30-12-21', '31-12-21', '01-01-22', '02-01-22', '03-01-22', '05-01-22', '06-01-22', '07-01-22', '08-01-22', '09-01-22', '10-01-22']
    chart = PControlChart(c=c, n=n, xlabel="x-label", ylabel="y-label")
    chart.dates = dates
    chart.dateformat = "%d-%m-%y"
    #stages=chart.stages(data=chart.c, max_stages=2)
    #if stages is not None:
    #    chart.split(stages)
    #chart.split([10])
    #chart.limits=True
    chart.append_rules([Rule01(), Rule02(), Rule03(), Rule04(), Rule05(), Rule06(), Rule07(), Rule08()])
    chart.plot()

    df1 = chart.data(0)
    #print(df1[["CL", "UCL", "LCL"]])
    #print("stable={0}".format(chart.stable()))



"""
Attribute charts: p chart is also known as the control chart for proportions.

https://sixsigmastudyguide.com/p-attribute-charts/
"""

import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import pandas as pd
from datetime import datetime
from SPC import ControlChart

class PControlChart(ControlChart):
    def __init__(self, c:list, n:list, xlabel:str="", ylabel:str=""):
        """ Initialization.

            :param c: number of defects in the samples.
            :param n: sample sizes.
            :param xlabel: x-as label.
            :param ylabel: y-as label.
        """
        # Initialization of the base class.
        super().__init__(1) # P chart.

        # Remember the parameters.
        self.number_of_defects = len(c)
        self._xlabel = xlabel
        self._ylabel = ylabel

        # Remember the number of defective in the sample.
        self.c = c

        # Remember the sample sizes.
        self.n = n

        # Initialize the arrays.
        self.value_p = np.zeros(((self.number_of_defects), 1))
        self.cl_p = np.zeros(((self.number_of_defects), 1))
        self.ucl_p = np.zeros(((self.number_of_defects), 1))
        self.lcl_p = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_min = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_min = np.zeros(((self.number_of_defects), 1))

        # Set the values.
        for i in range(self.number_of_defects):
            self.value_p[i] = c[i] / n[i]

        # Calculate the UCL, CL, LCL.
        self.cl_p[:] = self.value_p.mean() 
        for i in range(self.number_of_defects):
            self.ucl_p[i] = self.value_p.mean() + 3 * np.sqrt((self.value_p.mean() * (1 - self.value_p.mean()))/n[i])
            self.lcl_p[i] = self.value_p.mean() - 3 * np.sqrt((self.value_p.mean() * (1 - self.value_p.mean()))/n[i])

        # Calculate the one and two sigma.
        for i in range(self.number_of_defects):
            self.two_sigma_plus[i] = self.value_p.mean() + 3 * np.sqrt((self.value_p.mean() * (1 - self.value_p.mean()))/n[i]) * 2/3
            self.one_sigma_plus[i] = self.value_p.mean() + 3 * np.sqrt((self.value_p.mean() * (1 - self.value_p.mean()))/n[i]) * 1/3
            self.one_sigma_min[i] = self.value_p.mean() - 3 * np.sqrt((self.value_p.mean() * (1 - self.value_p.mean()))/n[i]) * 1/3
            self.two_sigma_min[i] = self.value_p.mean() - 3 * np.sqrt((self.value_p.mean() * (1 - self.value_p.mean()))/n[i]) * 2/3

    def plot(self):
        """ Create the P chart.
        """
        plt.figure(figsize=(15,5))

        # The x-axis can be numeric or datetime.
        if (len(super().dates) == 0):
            x_values_p = list(range(0, self.number_of_defects))
        else:
            format=super().dateformat
            x_values_p = [datetime.strptime(d, format).date() for d in super().dates]
            plt.gca().xaxis.set_major_formatter(mdates.DateFormatter(super().dateformat))

        # P chart.
        plt.plot(x_values_p, self.value_p, marker="o",color="k",label="p")
        plt.plot(x_values_p, self.ucl_p,color="r",label="UCL")

        # The limits indicator for +2s, +1s.
        if super().limits:
            plt.plot(x_values_p, self.two_sigma_plus, color="r", linestyle='dashed', label="+2s")
            plt.plot(x_values_p, self.one_sigma_plus, color="r", linestyle='dashed', label="+1s")

        plt.plot(x_values_p, self.cl_p,color="b",label="CL")

        # The limits indicator for -1s, -2s.
        if super().limits:
            plt.plot(x_values_p, self.one_sigma_min, color="r", linestyle='dashed', label="-1s")
            plt.plot(x_values_p, self.two_sigma_min, color="r", linestyle='dashed', label="-2s")

        plt.plot(x_values_p, self.lcl_p,color="r",label="LCL")
        plt.title("P Control Chart")

        # Add a legend.
        plt.legend(loc='upper right')

        # Set the x-label.
        plt.xlabel( self._xlabel)

        # Set the y-label.
        plt.ylabel( self._ylabel)

        plt.show()

    def split(self, stages:list):
        """ Split the chart.

            :param stages: The stages.
        """
        # Include the last index.
        if not self.number_of_defects in stages:
            stages.append(self.number_of_defects)

        # Initialize the arrays.
        self.cl_p = np.zeros(((self.number_of_defects), 1))
        self.ucl_p = np.zeros(((self.number_of_defects), 1))
        self.lcl_p = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_plus = np.zeros(((self.number_of_defects), 1))
        self.two_sigma_min = np.zeros(((self.number_of_defects), 1))
        self.one_sigma_min = np.zeros(((self.number_of_defects), 1))

        # Make the calculations on each split, instead of the full data.
        start_index = 0
        for i in stages:
            # Set the end index.
            end_index = i

            # Calculate the UCL, CL, LCL.
            self.cl_p[start_index:end_index] = self.value_p[start_index:end_index].mean()
            for i in range(start_index, end_index):
                self.ucl_p[i] = self.value_p[start_index:end_index].mean() + 3 * np.sqrt((self.value_p[start_index:end_index].mean() * (1 - self.value_p[start_index:end_index].mean()))/self.n[i])
                self.lcl_p[i] = self.value_p[start_index:end_index].mean() - 3 * np.sqrt((self.value_p[start_index:end_index].mean() * (1 - self.value_p[start_index:end_index].mean()))/self.n[i])

            # Calculate the one and two sigma.
            for i in range(start_index, end_index):
                self.two_sigma_plus[i] = self.value_p[start_index:end_index].mean() + 3 * np.sqrt((self.value_p[start_index:end_index].mean() * (1 - self.value_p[start_index:end_index].mean()))/self.n[i]) * 2/3
                self.one_sigma_plus[i] = self.value_p[start_index:end_index].mean() + 3 * np.sqrt((self.value_p[start_index:end_index].mean() * (1 - self.value_p[start_index:end_index].mean()))/self.n[i]) * 1/3
                self.one_sigma_min[i] = self.value_p[start_index:end_index].mean() - 3 * np.sqrt((self.value_p[start_index:end_index].mean() * (1 - self.value_p[start_index:end_index].mean()))/self.n[i]) * 1/3
                self.two_sigma_min[i] = self.value_p[start_index:end_index].mean() - 3 * np.sqrt((self.value_p[start_index:end_index].mean() * (1 - self.value_p[start_index:end_index].mean()))/self.n[i]) * 2/3

            # Set the start index.
            start_index = end_index

    def data(self, index:int):
        """ Returns the data.

            :param index: The index for the data (0 = P chart)
        """
        if index == 0: # P chart.
            df = pd.DataFrame(np.column_stack([self.value_p, self.ucl_p, self.two_sigma_plus, self.one_sigma_plus, self.cl_p, self.one_sigma_min, self.two_sigma_min, self.lcl_p]), columns=['value', 'UCL', '+2s', '+1s', 'CL', '-1s', '-2s', 'LCL'])
            self.execute_rules(df)

            # Check numerical or datetime for the x-axis.
            if (len(super().dates) != 0):
                df['date'] = super().dates
                df=df.set_index('date')

            return df

        raise ValueError

    def stable(self):
        """ Returns the stable indicator.
        """
        # Execute the rules.
        df = self.data(0)

        if True in df["SIGNAL"].values:
            return False

        return True



"""
The Process Capability chart.
"""

import numpy as np
import matplotlib.pyplot as plt
import statistics

class ProcessCapabilityChart():
    def __init__(self, data:list, target:int, LSL:int, USL:int):
        """ Initialization.

            :param data: The data.
            :param target: The target.
            :param LSL: The lower specification limit.
            :param USL: The upper specification limit.
        """
        # Remember the parameters.
        self._data=data
        self._target = target
        self._LSL = LSL
        self._USL = USL

        # Calculate Cp.
        self.Cp = (USL-LSL)/(6*np.std(data))

        # Calculate Cpk.
        self.Cpk = min((USL-statistics.mean(data))/(3*statistics.stdev(data)), (statistics.mean(data)-LSL)/(3*statistics.stdev(data)))

        # Get data summary statistics.
        self.num_samples = len(data)
        self.sample_mean = statistics.mean(data)
        self.sample_std = statistics.stdev(data)
        self.sample_max = max(data)
        self.sample_min = min(data)
        self.sample_median = statistics.median(data)

        # Get percentages outside the specification limits.
        self.pct_below_LSL = len([item for item in data if item < LSL])/self.num_samples*100
        self.pct_above_USL = len([item for item in data if item > USL])/self.num_samples*100
        self.reject_rate = self.pct_below_LSL + self.pct_above_USL

    def plot(self, bins:int):
        """ Create the plot.

            :param bins: The number of bins.
        """
        plt.figure(figsize=(15,10))
        plt.hist(self._data, bins=bins, color="lightgrey", edgecolor="black", density=False)
        plt.axvline(self._LSL, linestyle="dashed", color="red", label="LSL")
        plt.axvline(self._target, linestyle="dashed", color="grey", label="Target")
        plt.axvline(self._USL, linestyle="dashed", color="red", label="USL")
        plt.title("Histogram")
        plt.legend()
        plt.show()

    def capable(self, target_cpk:float):
        """ Returns the capable indicator.

            :param target_cpk: The target Cpk value.
        """
        return self.Cpk > target_cpk
